#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef DJGPP
#include <crt0.h>
#include <conio.h>
#else
#include <termio.h>
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include <sys/mman.h>

#include <errno.h>

/* GOALS: */
static unsigned xres;      /* pixels */
static unsigned yres;      /* pixels */
static unsigned refresh;   /* Hz */

/* PARAMETERS: */
static unsigned hsyncfreq; /* 30..70 kHz  (as Hz) */
static unsigned dispclock; /* 25..155 MHz (as Hz) */

static unsigned left=0,right=0, top=0,bottom=0;
static double vslength = 46e-6; /* vertical sync: 46 microseconds */
static double hslength =  1e-6; /* horizontal sync: 1 microsecond */

/* RESULTS: */
static unsigned hslen, vslen;

/* LIMITS: */
static unsigned mclk = 83e6;     /* Maximum xres*yres*refresh (Hz) */
static unsigned hsyncmin = 30e3; /* Minimum hsync freq (Hz) */
static unsigned hsyncmax = 70e3; /* Maximum hsync freq (Hz) */
static unsigned vsyncmin = 50;   /* Minimum refresh rate (Hz) */
static unsigned vsyncmax = 120;  /* Maximum refresh rate (Hz) */

static void GuessClocks(unsigned xblank, unsigned yblank)
{
    /* We want about 47 microseconds for the vertical sync.
     *
     * Length of (yres + yblank + yslen) = 1/refresh seconds.
     * 
     * Length of yslen is yslen*(1/refresh)/(yres+yblank+yslen)
     *
     * yslen*(1/refresh)/(yres+yblank+yslen) = 47e-6
     *
     * yslen = (yres+yblank) * 47e-6 / (1/refresh - 47e-6)
     */
    
    vslen = (unsigned) (ceil((yres+yblank) * vslength / (1.0/refresh - vslength)));
    
    /* hsyncfreq = (yslen + yblank + yres) * refreshfreq
     */
    
    hsyncfreq = (vslen + yblank + yres) * refresh;
    
    /* We want about 1 microseconds for the horizontal sync.
     * 
     * Length of (xres + xblank + xslen) = 1/hsyncfreq seconds.
     *
     * xslen = (xres+xblank) * 1e-6 / (1/hsyncfreq - 1e-6)
     */
    
    hslen = (unsigned) (ceil((xres+xblank) * hslength / (1.0/hsyncfreq - hslength)));
    
    dispclock = refresh * (xres+xblank+hslen) * (yres+yblank+vslen);
}

static void VerifyMode(void)
{
    unsigned pixrate;
    fflush(stdout);
    if(refresh < vsyncmin) { fprintf(stderr, "# Too low vsync (%u < %u)\n", refresh, vsyncmin); }
    if(refresh > vsyncmax) { fprintf(stderr, "# Too high vsync (%u > %u)\n", refresh, vsyncmax); }
    pixrate = xres * yres * refresh;
    if(pixrate > mclk) { fprintf(stderr, "# Too high pixel access rate (%u > %u)\n", pixrate, mclk); }
    if(hsyncfreq < hsyncmin) { fprintf(stderr, "# Too low hsync (%u < %u)\n", hsyncfreq, hsyncmin); }
    if(hsyncfreq > hsyncmax) { fprintf(stderr, "# Too high hsync (%u > %u)\n", hsyncfreq, hsyncmax); }
    
    if(1.0/hsyncfreq < hslength) { fprintf(stderr, "IMPOSSIBLE MODE: scanline is shorter (%.4fs) than hsync time (%.1f s)\n", 1e6/hsyncfreq, hslength*1e6); }
    if(1.0/refresh  < vslength) { fprintf(stderr, "IMPOSSIBLE MODE: screen update is shorter (%.4fs) than vsync time (%.1f s)\n", 1e6/refresh, vslength*1e6); }
    
    /* These errors only apply for atyfb */
    if(xres/8 - 1 > 0xFF)            { fprintf(stderr, "# Mode invalid for ATYFB: xres (%u) > 2048 (h_disp too large)\n", xres); }
    if(xres/8 - 1 + right/8 > 0x1FF) { fprintf(stderr, "# Mode invalid for ATYFB: xres+right (%u) > 4096 (h_sync_start too large)\n", xres+right); }
    if((hslen+7) / 8 > 0x1F)         { fprintf(stderr, "# Mode invalid for ATYFB: hslen (%u) > 248 (h_sync_wid too large)\n", hslen); }

    if(yres - 1 > 0x7FF)             { fprintf(stderr, "# Mode invalid for ATYFB: yres (%u) > 2048 (v_disp too large)\n", xres); }
    if(yres - 1 + bottom > 0x7FF)    { fprintf(stderr, "# Mode invalid for ATYFB: yres+bottom (%u) > 2048 (v_sync_strt too large)\n", yres+bottom); }
    if(vslen > 0x1F)                 { fprintf(stderr, "# Mode invalid for ATYFB: vslen (%u) > 31 (v_sync_wid too large)\n", vslen); }
}

static void DumpMode_fbset(void)
{
    fflush(stdout);
	fprintf(stderr,
	    "# fbset mode:\n"
		"mode \"%ux%u-%u\"\n"
		"    # D: %.3f MHz, H: %.3f kHz, V: %.3f Hz\n"
		"    geometry %u %u   %u %u   %u\n"
		"    timings %u  %u %u  %u %u  %u %u\n"
		"    vsync high\n"
		"    hsync high\n"
		"endmode\n",
		xres, yres, refresh,
		dispclock/1e6,
		hsyncfreq/1e3,
		refresh/1.0,
		xres, yres,  xres, yres,  8,
		(unsigned)(1e12 / dispclock),
		left, right,
		top, bottom,
		hslen, vslen);
}

static void DumpMode_xf86(void)
{
	fprintf(stderr,
		"# XF86Config modeline\n"
		"Modeline \"%ux%u-%u\"     %-8.3f %u %u %u %u   %u %u %u %u\n",
		xres, yres, refresh,
		dispclock / 1e6,
		xres, left+xres, left+xres+hslen, left+xres+hslen+right,
		yres, top+yres,  top+yres+vslen,  top+yres+vslen+bottom);
}

static void DumpMode_stm(void)
{
	fprintf(stderr,
		"# TextConfig modeline\n"
		"\"%ux%u-%uHz\"     %-8.3f %u %u %u %u   %u %u %u %u\n",
		xres, yres, refresh,
		dispclock / 1e6,
		xres, left+xres, left+xres+hslen, left+xres+hslen+right,
		yres, top+yres,  top+yres+vslen,  top+yres+vslen+bottom);
}

static unsigned LINES, COLS;

static void GetScreenGeometry(void)
{
#if defined(DJGPP) || defined(__BORLANDC__)
    struct text_info w;
    memset(&w, 0, sizeof(w));
    gettextinfo(&w);
    LINES=w.screenheight;
    COLS =w.screenwidth;
#else
#ifdef TIOCGWINSZ
    struct winsize w;
    memset(&w, 0, sizeof(w));
    if(ioctl(1, TIOCGWINSZ, &w) >= 0)
    {
        LINES=w.ws_row;
        COLS =w.ws_col;
    }
    else
#else
#ifdef WIOCGETD
    struct uwdata w;
    memset(&w, 0, sizeof(w));
    LINES = COLS = 0;
    if(ioctl(1, WIOCGETD, &w) >= 0)
    {
        LINES = w.uw_height / w.uw_vs;
        COLS  = w.uw_width / w.uw_hs;
    }
    else
#endif
#endif
    LINES = COLS = 0;
#endif
}

static void PutC(unsigned char *map, unsigned x, unsigned y, unsigned char colour, unsigned char mrk)
{
    map[(y*COLS+x)*2]   = mrk;
    map[(y*COLS+x)*2+1] = colour;
}

static void FillScreen(void)
{
    FILE *fp = fopen("/dev/vcsa0", "rb+");
    unsigned char *map;
    unsigned x, y, xma, yma;
    int hasmmap = 0;
    
    if(!fp)return;
    
    if(!LINES || !COLS)
    {
        fclose(fp);
        return;
    }

    map = mmap(NULL, LINES*COLS*2, PROT_WRITE, MAP_PRIVATE, fileno(fp), 4);
    
    if(map == MAP_FAILED)
    {
        /* Actually, the mmap *always* fails for vcs-devices :( */
        map = malloc(LINES*COLS*2);
        if(!map)
        {
            fclose(fp);
            return;
        }
    }
    else
    {
        hasmmap = 1;
        fclose(fp);
    }
    
    xma = COLS/6;  if(xma < 4)xma = 4;
    yma = LINES/6; if(yma < 4)yma = 4;
    
    for(y=0; y<LINES; ++y)
        for(x=0; x<COLS; ++x)
            PutC(map, x,y, 0x07, "PRESS A KEY  "[x%(6+2+3+2)]);

    for(y=0; y<yma; ++y)
        for(x=0; x<xma; ++x)
        {
            PutC(map, x,       y,         0x1F, '0'+(y>x?y:x)%10);
            PutC(map, COLS-x-1,y,         0x4F, '0'+(y>x?y:x)%10);
            PutC(map, COLS-x-1,LINES-y-1, 0x2E, '0'+(y>x?y:x)%10);
            PutC(map, x,       LINES-y-1, 0xD0, '0'+(y>x?y:x)%10);
        }
    
    if(hasmmap)
    {
        munmap(map, LINES*COLS*2);
    }
    else
    {
        fseek(fp, 4, SEEK_SET);
        fwrite(map, 2, LINES*COLS, fp);
        fclose(fp);
    }
}

static void SetMode(void)
{
    char Buf[512];
    
    GuessClocks(left+right, top+bottom);
    
    sprintf(Buf,
        "fbset -g %u %u %u %u 8 -t %u %u %u %u %u %u %u -vsync high -hsync high",
        xres, yres,
        xres, yres,
        (unsigned) (1e12 / dispclock),
        left, right,
        top, bottom,
        hslen, vslen
    );
    
    fprintf(stderr, "Executing: '%s'\n", Buf);
    system(Buf);
    
    printf("\33[%d;%dH", LINES-10, 1);
    VerifyMode();

    DumpMode_fbset();
    DumpMode_xf86();
    DumpMode_stm();
    
    GetScreenGeometry();
}

#include <argh.h>

static void PrintVersionInfo(void)
{
    printf("modetest - video mode composer v"VERSION" (C) 1992,2001 Bisqwit\n");
}

static void ArghValueError(long l)
{
    fprintf(stderr, "modetest: parameter value out of range: `%ld'\n", l);
}

static void ArghValueErrorS(const char *s)
{
    fprintf(stderr, "modetest: invalid value: `%s'\n", s);
}

int main(int argc, const char *const *argv)
{
    int Heelp = 0;
    argh_init();
    
    argh_add_long("help",   'h'); argh_add_bool('h'); argh_add_desc('h', "Help", NULL);
    argh_add_long("version",'V'); argh_add_bool('V'); argh_add_desc('V', "Version information", NULL);
    argh_add_long("xres",   'x'); argh_add_string('x', 1, 64); argh_add_desc('x', "Horizontal resolution.", "<visible>[+<invisible>[/<left%>]]");
    argh_add_long("yres",   'y'); argh_add_string('y', 1, 64); argh_add_desc('y', "Vertical resolution.", "<visible>[+<invisible>[/<top%>]]");
    argh_add_long("refresh",'r'); argh_add_int('r',1,65535); argh_add_desc('r', "Screen refresh rate.", "<num>");
    argh_add_long("hslen",  'X'); argh_add_float('X', 0.0, 1.0); argh_add_desc('X', "Horizontal sync length. Range: 0..1 s, default: 1e-6 (1 microsecond)", "<num>");
    argh_add_long("vslen",  'Y'); argh_add_float('Y', 0.0, 1.0); argh_add_desc('Y', "Vertical sync length. Range: 0..1 s, default: 46e-6 (46 microseconds)", "<num>");
    
    argh_start_parse(argc, argv);
    
    for(;;)
    {
        int c = argh_get_param();
        if(c == -1)break;
        switch(c)
        {
            case 'V':
                PrintVersionInfo();
                return 0;
            case 'h':
                Heelp = 1;
                break;
            case 'x':
            {
                const char *s = argh_get_string();
                unsigned blank;
                char *ep;
                long l = strtol(s, &ep, 10);
                if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                xres = l;
                if(*ep == '+')
                {
                    double perc = 80.0;
                    l = strtol(s=ep+1, &ep, 10);
                    if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                    blank = l;
                    if(*ep == '/')
                    {
                        errno = 0;
                        perc = strtod(ep+1, &ep);
                        if((ep && *ep) || perc<0 || perc>100.0) { ArghValueErrorS(ep); return -1; }
                    }
                    left = blank*perc/100.0;
                    right = blank-left;
                }
                break;
            }
            case 'y':
            {
                const char *s = argh_get_string();
                unsigned blank;
                char *ep;
                long l = strtol(s, &ep, 10);
                if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                yres = l;
                if(*ep == '+')
                {
                    double perc = 90.0;
                    l = strtol(s=ep+1, &ep, 10);
                    if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                    blank = l;
                    if(*ep == '/')
                    {
                        errno = 0;
                        perc = strtod(ep+1, &ep);
                        if((ep && *ep) || perc<0 || perc>100.0) { ArghValueErrorS(ep); return -1; }
                    }
                    top = blank*perc/100.0;
                    bottom = blank-top;
                }
                break;
            }
            case 'X':
                hslength = argh_get_float();
                break;
            case 'Y':
                vslength = argh_get_float();
                break;
            case 'r':
                refresh = argh_get_int();
                break;
            default:
                fprintf(stderr,
                    "modetest: Parameter `%s' is unrecognized.\n",
                    argh_get_string());
                return -1;
        }
    }
    if(!argh_ok())return -1;
    
    if(!xres || !yres || !refresh)
    {
        fprintf(stderr, "Missing options.\n");
        Heelp=1;
    }
    
    if(Heelp)
    {
        PrintVersionInfo();
        printf(
            "\nDevelopes a screen mode."
            "\n");
        printf(
            "Usage: modetest <options> [<...>]\n"
            "Options:\n");
        argh_list_options();
        printf(
        	"\n"
        	"You must specify at least xres, yres and refresh.\n"
        	"Current assumed monitor and display card limits:\n"
        	"    Memory clock: %.1f MHz\n"
        	"    Horizontal sync range: %u .. %u kHz\n"
        	"    Vertical sync (refresh) range: %u .. %u Hz\n",
        		mclk/1e6,
        		hsyncmin/1000, hsyncmax/1000,
        		vsyncmin, vsyncmax
        );
        printf(
        	"\n"
        	"The program tries to set the screenmode with fbset command.\n"
        	"You might want to redirect the stderr to a file in case the fbset command\n"
        	"fails and renders the screen garbled: modetest [<options>] 2>logfile.txt\n"
        	"\n"
        	"Example: modetest 2>tmptmp -x1280+296/78.4 -y1024+39/99 -r60\n"
        	"         (gives the screenmode I'm using with my framebuffer console)\n"
        );
        return 0;
    }
    
    FillScreen();
    SetMode();
    FillScreen();
    
    printf("\33[%d;%dH", LINES, 1);
    getchar();

    /* Restore the videomode to normal */
    system("fbset -g 1280 1024 1504 1024 8 -t 9292 232 64 38 1 112 3 -accel true");
    
    return 0;
}
